import logging
from collections import namedtuple
import json
from functools import wraps

from flask import request, Response as FlaskResponse
from werkzeug.exceptions import (
    HTTPException, BadRequest, InternalServerError, default_exceptions
)
from funcportal.function import InvalidArgumentsError, MissingArgumentsError


logger = logging.getLogger(__name__)


Response = namedtuple('Response', ['status_code', 'data'])


def _describe_arguments(arguments):
    return {
        'required': [
            {'name': arg.name} for arg in arguments if arg.required
        ],
        'optional': [
            {'name': arg.name, 'default': arg.default}
            for arg in arguments if not arg.required
        ]
    }


def handle_errors(handler_method):

    @wraps(handler_method)
    def wrapped(obj, *args, **kwargs):
        try:
            return handler_method(obj, *args, **kwargs)
        except InvalidArgumentsError:
            return Response(
                400, {'error': 'Invalid arguments were passed.'}
            )
        except MissingArgumentsError as e:
            data = {
                'error': str(e),
                'arguments': _describe_arguments(
                    obj.portal_function.arguments
                )
            }
            return Response(400, data)
        except Exception:
            logger.exception(
                'Error evaluating the function {}'
                .format(obj.portal_function.name)
            )
            return Response(500, {'error': 'Internal server error.'})

    return wrapped


class BaseHandler(object):

    def __init__(self, portal_function):
        self.portal_function = portal_function

    @handle_errors
    def _call_function(self, input_data):
        return Response(200, {'result': self.portal_function(**input_data)})


class BaseQueueHandler(BaseHandler):

    def __init__(self, portal_function, queue):
        super(BaseQueueHandler, self).__init__(portal_function)
        self.queue = queue

    @handle_errors
    def _submit_function(self, input_data):
        return Response(
            202,
            {'result_token':
                self.portal_function.submit(self.queue, **input_data)}
        )


def render_response(response, function_name):
    status_code = response.status_code
    try:
        body = json.dumps(response.data)
    except Exception:
        logger.exception(
            'Error serialising output from function {} as JSON'
            .format(function_name)
        )
        status_code = 500
        body = json.dumps({'error': 'Internal server error.'})
    return body, status_code


def render_flask_response(response, function_name):
    body, status = render_response(response, function_name)
    return FlaskResponse(body, status=status, mimetype='application/json')


class FlaskHandler(BaseHandler):

    def __call__(self):
        if self.portal_function.requires_arguments():
            try:
                inputs = request.get_json(force=True)
            except BadRequest:
                return render_flask_response(
                    Response(400, {'error': 'Malformatted JSON.'}),
                    self.portal_function.name
                )
        else:
            inputs = {}

        response = self._call_function(inputs)
        return render_flask_response(response, self.portal_function.name)


class FlaskQueueSubmissionHandler(BaseQueueHandler):

    def __call__(self):

        if self.portal_function.requires_arguments():
            try:
                inputs = request.get_json(force=True)
            except BadRequest:
                return render_flask_response(
                    Response(400, {'error': 'Malformatted JSON.'}),
                    self.portal_function.name
                )
        else:
            inputs = {}

        response = self._submit_function(inputs)
        return render_flask_response(response, self.portal_function.name)


class FlaskQueueRetrievalHandler(BaseQueueHandler):

    def __call__(self, result_token):

        job = self.queue.fetch_job(result_token)

        if job is None:
            response = Response(404, {'error': 'No such job.'})
        if job.result is None:
            response = Response(404, {'error': 'Job result not available.'})
        else:
            output_data = job.result
            response = Response(200, {'result': output_data})

        return render_flask_response(response, self.portal_function.name)


def configure_flask_app(app):
    for exception_type in default_exceptions.values():
        app.register_error_handler(exception_type, _flask_json_errorhandler)


def _flask_json_errorhandler(exception):
    """Create a JSON-encoded flask Response from an Exception."""

    if not isinstance(exception, HTTPException):
        exception = InternalServerError()

    body = json.dumps({
        'error': exception.name,
        'description': exception.description
    })
    status_code = exception.code

    return FlaskResponse(body, status=status_code, mimetype='application/json')



from collections import namedtuple
import six

if six.PY3:
    from inspect import signature, Parameter
else:
    from funcsigs import signature, Parameter


Argument = namedtuple('Argument', ['name', 'required', 'default'])


class MissingArgumentsError(TypeError):

    def __init__(self, function_name, missing_arguments):
        self.missing_arguments = missing_arguments
        message = '{}() missing {} required positional argument{}: {}'.format(
            function_name,
            len(missing_arguments),
            '' if len(missing_arguments) == 1 else 's',
            ', '.join(repr(arg) for arg in missing_arguments)
        )
        super(MissingArgumentsError, self).__init__(message)


class InvalidArgumentsError(Exception):
    pass


def _get_arguments(function):

    sig = signature(function)

    arguments = []

    for parameter in sig.parameters.values():
        required = parameter.default is Parameter.empty
        arguments.append(Argument(parameter.name, required, parameter.default))

    return arguments


class PortalFunction(object):

    def __init__(self, function):
        self.function = function
        self.name = self.function.__name__
        self.arguments = _get_arguments(function)

    def requires_arguments(self):
        return any(arg.required for arg in self.arguments)

    def _check_arguments(self, **kwargs):
        required = set(arg.name for arg in self.arguments if arg.required)
        provided = set(kwargs.keys())

        missing = required - provided
        extra = provided - required

        if missing:
            raise MissingArgumentsError(self.name, missing)
        if extra:
            # TODO: Consider logging this case
            pass

    def __call__(self, **kwargs):
        self._check_arguments(**kwargs)
        return self.function(**kwargs)

    def submit(self, queue, **kwargs):
        self._check_arguments(**kwargs)
        job = queue.enqueue(self.function, **kwargs)
        return job.id



from flask import Flask
from redis import Redis
from rq import Queue, Worker

from funcportal.function import PortalFunction
from funcportal.handler import (
    FlaskHandler, FlaskQueueSubmissionHandler, FlaskQueueRetrievalHandler,
    configure_flask_app
)


REDIS_CONNECTION = Redis()
JOB_QUEUE = Queue('funcportal', connection=REDIS_CONNECTION)


def run_async_worker():
    worker = Worker([JOB_QUEUE], connection=REDIS_CONNECTION)
    worker.work()


class Portal(object):

    def __init__(self):
        self.routes = {}
        self._queue = None

    def register_endpoint(self, endpoint, function, asynchronous=False):
        self.routes[endpoint] = (PortalFunction(function), asynchronous)

    def generate_wsgi_app(self):
        app = Flask('funcportal')
        configure_flask_app(app)
        for endpoint, (function, asynchronous) in self.routes.items():
            if asynchronous:
                handler = FlaskQueueSubmissionHandler(function, JOB_QUEUE)
            else:
                handler = FlaskHandler(function)
            app.add_url_rule(
                endpoint, endpoint, handler, methods=['POST']
            )

            if asynchronous:
                result_endpoint = endpoint.rstrip('/') + '/<result_token>'
                result_handler = FlaskQueueRetrievalHandler(
                    function, JOB_QUEUE
                )
                app.add_url_rule(
                    result_endpoint, result_endpoint, result_handler,
                    methods=['GET']
                )

        return app

    def run_wsgi(self):
        # TODO: Use gunicorn
        wsgi_app = self.generate_wsgi_app()
        wsgi_app.run()



import os
import sys


class FunctionImportError(Exception):
    pass


def import_function(module, function):
    """Import a function from a module."""

    try:
        __import__(module)
    except ImportError:
        if module.endswith('.py') and os.path.exists(module):
            tpl = "Failed to find function, did you mean '{}:{}'?"
            raise ImportError(tpl.format(module.rstrip('.py'), function))
        else:
            raise

    mod = sys.modules[module]

    try:
        obj = getattr(mod, function)
    except AttributeError:
        tpl = 'Failed to find function {!r} in module {!r}'
        raise FunctionImportError(tpl.format(function, module))

    if not callable(obj):
        raise FunctionImportError('Object is not callable: {!r}'.format(obj))

    return obj



import argparse
from collections import namedtuple

import yaml

from funcportal.app import Portal, run_async_worker
from funcportal import util


ERROR_TEMPLATE = "{!r} does not match format 'module:function[:endpoint]'"


Route = namedtuple('Route', ['module', 'function', 'endpoint', 'asynchronous'])


class ConfigurationError(RuntimeError):
    pass


def routes_from_config(path):

    with open(path) as fp:
        config = yaml.load(fp)

    try:
        route_data = config['routes']
    except KeyError:
        raise ConfigurationError(
            'No routes entry in config read from {}'.format(path)
        )

    routes = []

    for entry in route_data:
        try:
            route = Route(
                entry['module'], entry['function'], entry['endpoint'],
                entry.get('async', False)
            )
            routes.append(route)
        except (TypeError, KeyError):
            raise ConfigurationError('Malformed endpoint definition')

    return routes


def parse_route(arg):

    parts = arg.split(':')

    try:
        module = parts[0]
        function = parts[1]
    except IndexError:
        raise argparse.ArgumentTypeError(ERROR_TEMPLATE.format(arg))

    if len(module) == 0 or len(function) == 0:
        raise argparse.ArgumentTypeError(ERROR_TEMPLATE.format(arg))

    try:
        route = parts[2]
    except IndexError:
        route = function

    if not route.startswith('/'):
        route = '/' + route

    return Route(module, function, route, False)


def main():

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest='command')

    server = subparsers.add_parser('server')
    server.add_argument(
        'routes', nargs='+', type=parse_route,
        help='One or more route definitions, in the format ' +
             'module:function[:endpoint]. If not specified, the route will ' +
             'be the function name.'
    )
    server.add_argument(
        '--config', '-c', help='A YAML configuration file to load'
    )

    subparsers.add_parser('worker')

    args = parser.parse_args()

    if args.command == 'server':

        app = Portal()

        routes = []
        if args.config is not None:
            routes += routes_from_config(args.config)
        routes += args.routes

        for route in routes:
            function = util.import_function(route.module, route.function)
            app.register_endpoint(route.endpoint, function, route.asynchronous)

        app.run_wsgi()

    elif args.command == 'worker':
        run_async_worker()



import os
from setuptools import setup

from funcportal.version import __version__


def readme():
    source_root = os.path.dirname(__file__)
    path = os.path.join(source_root, 'README.rst')
    with open(path) as fp:
        return fp.read()


setup(
    name='funcportal',
    version=__version__,
    description='Serve Python functions as web APIs',
    long_description=readme(),
    packages=['funcportal'],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest',
    ],
    install_requires=[
        'flask',
        'requests',
        'six',
        'pyyaml',
        'rq',
        'redis'
    ],
    entry_points={
        'console_scripts': [
            'funcportal=funcportal.cli:main'
        ]
    }
)



import os
import subprocess
import time

import requests
import pytest


TEST_MODULE = """
def multiply(x, y):
    return x * y

def exponent(base, power=2):
    return base ** power
"""

TEST_CONFIG = """
routes:
  - module: module
    function: multiply
    endpoint: /fromconfig/multiply
  - module: module
    function: multiply
    endpoint: /async/multiply
    async: true
"""


@pytest.fixture(scope='session')
def srcdir(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp('tmp')
    tmpdir.join('module.py').write(TEST_MODULE)
    return tmpdir


@pytest.fixture(scope='session')
def server(srcdir):
    config = srcdir.join('portal.yaml')
    config.write(TEST_CONFIG)
    env = os.environ.copy()
    env['PYTHONPATH'] = os.environ.get('PYTHONPATH', '') + ':' + str(srcdir)
    process = subprocess.Popen(
        ['funcportal', 'server', 'module:multiply', 'module:exponent',
         '--config', str(config)],
        env=env
    )
    time.sleep(1)  # TODO: Check for readiness
    assert process.poll() is None, 'The funcportal process failed'
    yield 'http://127.0.0.1:5000'
    process.terminate()
    process.wait()


@pytest.fixture(scope='session')
def worker(srcdir):
    env = os.environ.copy()
    env['PYTHONPATH'] = os.environ.get('PYTHONPATH', '') + ':' + str(srcdir)
    process = subprocess.Popen(['funcportal', 'worker'], env=env)
    yield
    process.terminate()
    process.wait()


def test_funcportal(server):
    url = server + '/multiply'
    response = requests.post(url, json={'x': 2, 'y': 3})
    assert response.json() == {'result': 6}


def test_optional_arguments(server):
    url = server + '/exponent'
    response = requests.post(url, json={'base': 3, 'power': 4})
    assert response.json() == {'result': 81}


def test_optional_argument_omitted(server):
    url = server + '/exponent'
    response = requests.post(url, json={'base': 3})
    assert response.json() == {'result': 9}


def test_missing_argument(server):
    url = server + '/multiply'
    response = requests.post(url, json={'x': 2})
    assert response.status_code == 400
    body = response.json()
    assert "missing 1 required positional argument: 'y'" in body['error']
    assert body['arguments'] == {
        'required': [{'name': 'x'}, {'name': 'y'}],
        'optional': []
    }


def test_asynchronous(server, worker):
    url = server + '/async/multiply'
    response = requests.post(url, json={'x': 2, 'y': 3})
    assert response.status_code == 202
    result_token = response.json()['result_token']

    for _ in range(10):
        time.sleep(0.1)
        response = requests.get(url + '/{}'.format(result_token))
        print(response.text)
        if response.status_code != 404:
            break
    else:
        pytest.fail('result did not become available')

    assert response.status_code == 200
    assert response.json() == {'result': 6}
